import struct


def decode(fig, blob_id, scale, style_override_table):
    network = bytes(fig["blobs"][blob_id]["bytes"])

    i = 0
    num_vertices = struct.unpack("<I", network[i : i + 4])[0]
    i += 4
    num_segments = struct.unpack("<I", network[i : i + 4])[0]
    i += 4
    num_regions = struct.unpack("<I", network[i : i + 4])[0]
    i += 4

    vertices = []
    for vertex in range(num_vertices):
        # Should include stroke cap/limit, corner radius, mirroring, etc.
        style_id = struct.unpack("<I", network[i : i + 4])[0]

        # Coordinates
        x, y = struct.unpack("<ff", network[i + 4 : i + 12])

        vertices.append(decode_vertex(x, y, scale, style_override_table, style_id))
        i += 12

    segments = []
    for segment in range(num_segments):
        # No idea what it's for
        style_id = struct.unpack("<I", network[i : i + 4])[0]

        # Start vertex + tangent vector
        v1 = struct.unpack("<I", network[i + 4 : i + 8])[0]
        t1x, t1y = struct.unpack("<ff", network[i + 8 : i + 16])

        # End vertex + tangent vector
        v2 = struct.unpack("<I", network[i + 16 : i + 20])[0]
        t2x, t2y = struct.unpack("<ff", network[i + 20 : i + 28])

        segments.append(decode_segment(v1, v2, t1x, t1y, t2x, t2y, scale))
        i += 28

    regions = []
    for region in range(num_regions):
        # Flags should include winding rule
        flags, num_loops = struct.unpack("<II", network[i : i + 8])
        winding_rule = "NONZERO" if flags % 2 else "ODD"
        style_id = flags >> 1
        i += 8

        loops = []
        for loop in range(num_loops):
            num_loop_vertices = struct.unpack("<I", network[i : i + 4])[0]
            i += 4

            loop_vertices = []
            for vertex in range(num_loop_vertices):
                loop_vertices.append(struct.unpack("<I", network[i : i + 4])[0])
                i += 4
            loops.append(loop_vertices)

        regions.append(
            {
                "loops": loops,
                "style": style_override_table.get(style_id, {"styleID": style_id}),
                "windingRule": winding_rule,
            }
        )

    return {"regions": regions, "segments": segments, "vertices": vertices}


def decode_vertex(x, y, scale, style_override_table=None, style_id=None):
    vertex = {
        "x": 0 if (x == 0 or scale["x"] == 0) else x / scale["x"],
        "y": 0 if (y == 0 or scale["y"] == 0) else y / scale["y"],
    }

    if style_id:
        vertex["style"] = style_override_table.get(style_id, {"styleID": style_id})

    return vertex


def decode_segment(v1, v2, t1x, t1y, t2x, t2y, scale):
    return {
        "start": v1,
        "end": v2,
        "tangentStart": decode_vertex(t1x, t1y, scale),
        "tangentEnd": decode_vertex(t2x, t2y, scale),
    }



from .common import Point
from dataclasses import dataclass, field
from enum import IntEnum
from typing import Optional


class OverlayBackgroundInteraction(IntEnum):
    NONE = 0
    CLOSES_OVERLAY = 1


class PresentationStyle(IntEnum):
    SCREEN = 0
    OVERLAY = 1


class AnimationType(IntEnum):
    NONE = -1
    SLIDE_FROM_RIGHT = 0
    SLIDE_FROM_LEFT = 1
    SLIDE_FROM_BOTTOM = 2
    SLIDE_FROM_TOP = 3


@dataclass(kw_only=True)
class FlowOverlaySettings:
    _class: str = field(default="MSImmutableFlowOverlaySettings")
    overlayAnchor: Point
    sourceAnchor: Point
    offset: Point = field(default_factory=lambda: Point(0, 0))
    overlayType: int = 0

    @staticmethod
    def Positioned(position: str, offset: Point = Point(0, 0)) -> "FlowOverlaySettings":
        anchor = Point(0.5, 0.5)

        match position:
            case "TOP_LEFT":
                anchor = Point(0, 0)
            case "TOP_CENTER":
                anchor = Point(0.5, 0)
            case "TOP_RIGHT":
                anchor = Point(1, 0)
            case "BOTTOM_LEFT":
                anchor = Point(0, 1)
            case "BOTTOM_CENTER":
                anchor = Point(0.5, 1)
            case "BOTTOM_RIGHT":
                anchor = Point(1, 1)
            case "MANUAL":
                anchor = Point(0, 0)

        return FlowOverlaySettings(overlayAnchor=anchor, sourceAnchor=anchor, offset=offset)

    @staticmethod
    def RegularArtboard() -> "FlowOverlaySettings":
        anchor = Point(0.5, 0.5)

        return FlowOverlaySettings(overlayAnchor=anchor, sourceAnchor=anchor)


@dataclass(kw_only=True)
class FlowConnection:
    _class: str = field(default="MSImmutableFlowConnection")
    destinationArtboardID: str
    overlaySettings: Optional[FlowOverlaySettings]
    animationType: AnimationType = AnimationType.NONE
    maintainScrollPosition: bool = False
    shouldCloseExistingOverlays: bool = False


@dataclass(kw_only=True)
class PrototypeViewport:
    _class: str = field(default="MSImmutablePrototypeViewport")
    name: str
    size: Point
    # libraryID: str = 'EB972BCC-0467-4E50-998E-0AC5A39517F0'
    # templateID: str = '55992B99-92E5-4A93-AF90-B3A461675C05'



import copy
from . import base, group
from .context import context
from .config import config
from converter import utils
from sketchformat.layer_group import SymbolInstance, OverrideValue
from sketchformat.style import Style
from typing import List, Tuple
from .errors import Fig2SketchNodeChanged


def convert(fig_instance):
    if utils.is_invalid_ref(fig_instance["symbolData"]["symbolID"]):
        # Broken instance, return a placeholder in its place
        utils.log_conversion_warning("SYM001", fig_instance)
        return group.convert(fig_instance)

    all_overrides = get_all_overrides(fig_instance)
    sketch_overrides, unsupported = convert_overrides(all_overrides, fig_instance)
    if unsupported:
        if config.can_detach:
            utils.log_conversion_warning("SYM003", fig_instance, props=unsupported)

            # Modify input tree in place, with the detached symbol subtree
            detach_symbol(fig_instance, all_overrides)

            # Raise an exception to trigger conversion of the detached node
            raise Fig2SketchNodeChanged()
        else:
            utils.log_conversion_warning("SYM002", fig_instance, props=unsupported)

    # Use always the GUID of the master for the symbolID
    # The instance symbolID can refer to the overrideKey instead
    fig_master = context.find_symbol(fig_instance["symbolData"]["symbolID"])
    obj = SymbolInstance(
        **base.base_styled(fig_instance),
        symbolID=utils.gen_object_id(fig_master["guid"]),
        overrideValues=sketch_overrides,
    )
    # Replace style
    obj.style = Style(do_objectID=utils.gen_object_id(fig_instance["guid"], b"style"))

    return obj


def post_process(fig_instance, sketch_instance):
    if sketch_instance._class == "group":
        return group.post_process_frame(fig_instance, sketch_instance)
    else:
        return sketch_instance


def master_instance(fig_symbol):
    obj = SymbolInstance(
        **base.base_styled(fig_symbol),
        symbolID=utils.gen_object_id(fig_symbol["guid"]),
    )
    obj.do_objectID = utils.gen_object_id(fig_symbol["guid"], b"master_instance")

    # Replace style
    obj.style = Style(
        do_objectID=utils.gen_object_id(fig_symbol["guid"], b"master_instance_style")
    )

    return obj


def convert_overrides(all_overrides, fig_instance):
    sketch_overrides = []
    unsupported_overrides = []
    for override in all_overrides:
        sk, us = convert_override(override, fig_instance)
        sketch_overrides += sk
        unsupported_overrides += us

    return sketch_overrides, unsupported_overrides


def get_all_overrides(fig_instance):
    """Gets all overrides of a symbol, including component assignments"""

    # Convert top-level properties to overrides
    fig_master = context.find_symbol(fig_instance["symbolData"]["symbolID"])
    all_overrides = convert_properties_to_overrides(
        fig_master, fig_instance.get("componentPropAssignments", [])
    )

    # Sort overrides by length of path. This ensures top level overrides are processed before
    # nested ones which is required because a top override may change the symbol instance that is
    # used by child overrides
    for override in sorted(
        fig_instance["symbolData"]["symbolOverrides"],
        key=lambda x: len(x["guidPath"]["guids"]),
    ):
        guid_path = override["guidPath"]["guids"]
        new_override = {"guidPath": override["guidPath"]}
        for prop, value in override.items():
            if prop == "componentPropAssignments":
                nested_master = find_symbol_master(fig_master, guid_path, all_overrides)
                all_overrides += convert_properties_to_overrides(nested_master, value, guid_path)
            else:
                new_override[prop] = value

        all_overrides.append(new_override)

    # Do a pass to eliminate duplicate overrides, we use an ordered dict to keep them sorted
    unique_overrides = []
    for ov in all_overrides:
        guid = ov["guidPath"]["guids"]
        existing = [i for i in unique_overrides if i["guidPath"]["guids"] == guid]
        if existing:
            # Add properties to the previous override. Priority goes to the first item
            # because we want to prioritize prop assignments which we always convert first
            for k, v in ov.items():
                if k == "guidPath":
                    continue
                if k not in existing[0]:
                    existing[0][k] = v
        else:
            unique_overrides.append(ov)

    return unique_overrides


def convert_override(override: dict, fig_instance: dict) -> Tuple[List[OverrideValue], List[str]]:
    sketch_overrides = []
    unsupported_overrides = []

    try:
        # Convert uuids in the path from top symbol to child instance
        sketch_path = [
            utils.gen_object_id(context.fig_node(guid)["guid"])
            for guid in override["guidPath"]["guids"]
        ]
        sketch_path_str = "/".join(sketch_path)
    except KeyError as e:
        # Cannot find where to apply override
        utils.log_conversion_warning("SYM004", fig_instance, node_ref=e.args[0])
        return [], []

    for prop, value in override.items():
        if prop == "guidPath":
            continue
        if prop == "textData":
            # Text override.
            if "styleOverrideTable" in value:
                unsupported_overrides.append("textData.styleOverrideTable")
                continue

            sketch_overrides.append(
                OverrideValue(
                    overrideName=f"{sketch_path_str}_stringValue",
                    value=value["characters"],
                )
            )
        elif prop == "overriddenSymbolID":
            sketch_overrides.append(
                OverrideValue(
                    overrideName=f"{sketch_path_str}_symbolID",
                    value=utils.gen_object_id(value),
                )
            )
        elif prop in ["size", "pluginData", "name", "exportSettings"]:
            # Size is handled by applying derivedSymbolData
            # The rest are surely not worth detaching for
            pass
        else:
            # Unknown override
            unsupported_overrides.append(prop)

    return sketch_overrides, unsupported_overrides


def find_symbol_master(root_symbol, guid_path, overrides):
    current_symbol = root_symbol
    path = []
    for guid in guid_path:
        path.append(guid)
        # See if we have overriden the symbol_id
        symbol_id = [
            o["overriddenSymbolID"]
            for o in overrides
            if o["guidPath"]["guids"] == path and "overriddenSymbolID" in o
        ]
        if symbol_id:
            symbol_id = symbol_id[0]
        else:
            # Otherwise, find the instance
            instance = context.fig_node(guid)
            symbol_id = instance["symbolData"]["symbolID"]

        current_symbol = context.find_symbol(symbol_id)

    return current_symbol


def convert_properties_to_overrides(fig_master, properties, guid_path=[]):
    """Convert .fig property assignments to overrides.
    This makes it easier to work with them in a unified way."""
    overrides = []

    for prop in properties:
        for ref_prop, ref_guid in find_refs(fig_master, prop["defID"]):
            if ref_prop["componentPropNodeField"] == "OVERRIDDEN_SYMBOL_ID":
                override = {"overriddenSymbolID": prop["value"]["guidValue"]}
            elif ref_prop["componentPropNodeField"] == "TEXT_DATA":
                override = {"textData": prop["value"]["textValue"]}
            elif ref_prop["componentPropNodeField"] == "VISIBLE":
                override = {"visible": prop["value"]["boolValue"]}
            else:  # INHERIT_FILL_STYLE_ID
                raise Exception(f"Unexpected property {ref_prop['componentPropNodeField']}")

            overrides.append({**override, "guidPath": {"guids": guid_path + [ref_guid]}})

    return overrides


def find_refs(node, ref_id):
    """Find all usages of a property in a symbol, recursively"""
    refs = [
        (ref, node.get("overrideKey", node["guid"]))
        for ref in node.get("componentPropRefs", [])
        if ref["defID"] == ref_id and not ref.get("isDeleted", False)
    ]

    for ch in node.get("children", []):
        refs += find_refs(ch, ref_id)

    return refs


def detach_symbol(fig_instance, all_overrides):
    # Find symbol master
    fig_master = context.fig_node(fig_instance["symbolData"]["symbolID"])
    detached_children = copy.deepcopy(fig_master["children"], {})

    # Apply overrides to children
    for c in detached_children:
        apply_overrides(c, fig_instance["guid"], all_overrides, fig_instance["derivedSymbolData"])

    fig_instance["children"] = detached_children
    fig_instance["type"] = "FRAME"
    fig_instance["f2s_detached"] = True


def apply_overrides(fig_node, instance_id, overrides, derived_symbol_data):
    guid = fig_node.get("overrideKey", fig_node["guid"])

    # Apply overrides
    child_overrides = []
    for override in overrides:
        guids = override["guidPath"]["guids"]
        if guids[0] != guid:
            continue
        if len(guids) > 1:
            child_overrides.append({**override, "guidPath": {"guids": guids[1:]}})
        else:
            for k, v in override.items():
                if k == "guidPath":
                    continue
                elif k == "overriddenSymbolID":
                    fig_node["symbolData"]["symbolID"] = v
                else:
                    fig_node[k] = v

    # Recalculate size
    child_derived_data = []
    for derived in derived_symbol_data:
        guids = derived["guidPath"]["guids"]
        if guids[0] != guid:
            continue
        if len(guids) > 1:
            child_derived_data.append({**derived, "guidPath": {"guids": guids[1:]}})
        else:
            if "size" in derived:
                fig_node["size"] = derived["size"]
            if "transform" in derived:
                fig_node["transform"] = derived["transform"]
            # Should we do all properties instead?
            # A lot of them can be potentially set by derived data
            # by things like scaling the instance
            if "strokeWeight" in derived:
                fig_node["strokeWeight"] = derived["strokeWeight"]

    # Generate a unique ID by concatenating instance_id + node_id
    fig_node["guid"] = tuple(j for i in (instance_id, fig_node["guid"]) for j in i)

    # If it's an instance, pass the overrides down. Otherwise, convert the children
    if fig_node["type"] == "INSTANCE":
        fig_node["symbolData"]["symbolOverrides"] += child_overrides
        fig_node["derivedSymbolData"] += child_derived_data
    else:
        if fig_node.get("f2s_detached"):
            # This was previously an instance that was detached before this usage
            # We didn't change the override keys of children, and they are still relative
            # to the symbol, so we have to pass the key without our guid prefix
            overrides = child_overrides
            derived_symbol_data = child_derived_data

        for c in fig_node.get("children", []):
            apply_overrides(c, instance_id, overrides, derived_symbol_data)



#!/usr/bin/env python3

import argparse
import json
import logging
from zipfile import ZipFile, ZIP_DEFLATED, ZIP_STORED
import ssl
import sys
from typing import List

try:
    from importlib.metadata import version

    VERSION = version("fig2sketch")
except:
    VERSION = "unknown version"


def parse_args(args: List[str] = sys.argv[1:]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Converts a .fig document to .sketch")
    parser.add_argument("fig_file")
    parser.add_argument("sketch_file")

    group = parser.add_argument_group("conversion options")
    group.add_argument(
        "--instance-override",
        choices=["detach", "ignore"],
        default="detach",
        help="what to do when converting unsupported instance override (default = detach)",
    )
    group.add_argument(
        "--force-convert-images",
        action="store_true",
        help="try to convert corrupted images",
    )
    group.add_argument(
        "--compress",
        action="store_true",
        help="compress the output sketch file",
    )

    group = parser.add_argument_group("debug options")
    group.add_argument(
        "-v",
        action="count",
        dest="verbosity",
        help="return more details, can be repeated",
    )
    group.add_argument("--salt", type=str, help="salt used to generate ids, defaults to random")
    group.add_argument(
        "--dump-fig-json",
        type=argparse.FileType("w"),
        help="output a fig representation in json for debugging purposes",
    )

    parser.add_argument("--version", action="version", version=f"%(prog)s {VERSION}")

    return parser.parse_args(args)


def run(args: argparse.Namespace) -> None:
    # Set default log level
    level = logging.WARNING
    if args.verbosity:
        level = logging.INFO if args.verbosity == 1 else logging.DEBUG

    logging.basicConfig(level=level)

    # Import these after setting the log level
    from figformat import fig2tree
    from converter import convert
    from converter.config import config

    config.version = VERSION
    if args.salt:
        config.salt = args.salt.encode("utf8")

    if args.force_convert_images:
        from PIL import ImageFile

        ImageFile.LOAD_TRUNCATED_IMAGES = True

    config.can_detach = args.instance_override == "detach"

    # Load SSL certificates in OSs where Python does not use system defaults
    if not ssl.create_default_context().get_ca_certs():
        import certifi
        import os

        os.environ["SSL_CERT_FILE"] = certifi.where()
        logging.debug("Loaded TLS certificates from certifi")
    else:
        logging.debug("Using system TLS certificates")

    logging.debug(config)
    logging.debug(f"Version {VERSION}")

    compression = ZIP_STORED
    if args.compress:
        compression = ZIP_DEFLATED

    with ZipFile(args.sketch_file, "w", compression=compression) as output:
        fig_tree, id_map = fig2tree.convert_fig(args.fig_file, output)

        if args.dump_fig_json:
            json.dump(
                fig_tree,
                args.dump_fig_json,
                indent=2,
                ensure_ascii=False,
                default=lambda x: x.tolist(),
            )

        convert.convert_fig_tree_to_sketch(fig_tree, id_map, output)


def main() -> None:
    run(parse_args())


if __name__ == "__main__":
    main()



import logging
from . import component, page, font
from sketchformat.document import Swatch
from typing import Sequence, Tuple, Optional, Dict, IO, List


def find_symbols(node: dict) -> List[Sequence[int]]:
    if node["type"] == "SYMBOL":
        return [node["guid"]]

    found = []
    for child in node.get("children", []):
        found += find_symbols(child)

    return found


class Context:
    def init(self, components_page: Optional[dict], id_map: Dict[Sequence[int], dict]) -> None:
        self._sketch_components: Dict[Sequence[int], Swatch] = {}
        self.symbols_page = None
        self._node_by_id = id_map
        self._used_fonts: Dict[Tuple[str, str], Tuple[IO[bytes], str]] = {}
        self._component_symbols = (
            {s: False for s in find_symbols(components_page)} if components_page else {}
        )

        # Where to position symbols of the specified width
        # width -> (x, y)
        self._symbol_position = {0: [0, 0]}

    def component(self, cid: Sequence[int]) -> Tuple[dict, Optional[Swatch]]:
        fig_component = self.fig_node(cid)

        # See if we can convert this component to a Sketch swatch
        sketch_component = self._sketch_components.get(cid)
        if not sketch_component:
            sketch_component = component.convert(fig_component)
            if sketch_component is not None:
                self._sketch_components[cid] = sketch_component

        return fig_component, sketch_component

    def sketch_components(self) -> List[Swatch]:
        return list(self._sketch_components.values())

    def add_symbol(self, sketch_symbol):
        if not self.symbols_page:
            self.symbols_page = page.symbols_page()

        self.symbols_page.layers.append(sketch_symbol)
        self._position_symbol(sketch_symbol)

    def fig_node(self, fid: Sequence[int]) -> dict:
        return self._node_by_id[fid]

    def record_font(self, fig_font_name):
        font_descriptor = (fig_font_name["family"], fig_font_name["style"])
        font_info = self._used_fonts.get(font_descriptor)
        if font_info:
            return font_info[1]

        try:
            font_file, font_name = font.get_webfont(*font_descriptor)
        except:
            logging.warning(f"Could not download font {font_descriptor}")
            font_file = None
            if fig_font_name["postscript"]:
                font_name = fig_font_name["postscript"]
            else:
                font_name = f"{fig_font_name['family']}-{fig_font_name['style']}"

        self._used_fonts[font_descriptor] = (font_file, font_name)
        return font_name

    def used_fonts(self) -> Dict[Tuple[str, str], Tuple[IO[bytes], str]]:
        return self._used_fonts

    def find_symbol(self, sid: Sequence[int]) -> dict:
        symbol = self.fig_node(sid)
        sid = symbol["guid"]

        if not self._component_symbols.get(sid, True):
            # The symbol is in the component page and has not been converted yet, do it now
            from . import tree

            tree.convert_node(symbol, "")
            self._component_symbols[sid] = True

        return symbol

    def _position_symbol(self, sketch_symbol):
        # Mimics Sketch positioning algorith:
        #   All symbols of the same width go in the same column
        #   Each different width goes into a new column
        frame = sketch_symbol.frame
        width = frame.width

        position = self._symbol_position.get(width)
        if position:
            # Already have a column for this width, add it at the bottom
            frame.x = position[0]
            frame.y = position[1]
            position[1] += frame.height + 100
        else:
            # Create a new column at the end
            [last_width, [x, _]] = max(self._symbol_position.items(), key=lambda item: item[1][0])
            new_x = x + last_width + 100
            frame.x = new_x
            frame.y = 0
            self._symbol_position[width] = [new_x, frame.height + 100]


context = Context()



import hashlib
import logging
import struct
import uuid
from .config import config
from typing import Sequence, Dict, Optional

issued_warnings: Dict[tuple[int, int], list[str]] = {}


def gen_object_id(fig_id: Sequence[int], suffix: bytes = b"") -> str:
    # Generate UUIDs by hashing the fig GUID with a salt
    salted_id = config.salt + struct.pack("<" + "I" * len(fig_id), *fig_id) + suffix
    uuid_bytes = bytearray(hashlib.shake_128(salted_id).digest(16))

    # Override bits to match UUIDv4
    uuid_bytes[6] = (uuid_bytes[6] & 0x0F) | 0x40
    uuid_bytes[8] = (uuid_bytes[8] & 0x3F) | 0x80

    return str(uuid.UUID(bytes=bytes(uuid_bytes))).upper()


def generate_file_ref(data: bytes) -> str:
    return hashlib.sha1(hashlib.sha1(data).digest()).hexdigest()


def get_style_table_override(fig_item):
    override_table = {0: {}}

    if "styleOverrideTable" in fig_item:
        override_table.update({s["styleID"]: s for s in fig_item["styleOverrideTable"]})

    return override_table


UINT_MAX = 4294967295


def is_invalid_ref(uuid: Optional[Sequence[int]]) -> bool:
    return uuid is None or uuid[0] == UINT_MAX or uuid[1] == UINT_MAX


def has_rounded_corners(fig: dict) -> bool:
    return any(
        [
            fig.get(c, 0) != 0
            for c in [
                "rectangleTopLeftCornerRadius",
                "rectangleTopLeftCornerRadius",
                "rectangleTopLeftCornerRadius",
                "rectangleTopLeftCornerRadius",
            ]
        ]
    )


# Commented messages are no longer used, but we keep them for referencing past errors
WARNING_MESSAGES = {
    "TXT001": "is missing the glyphs property. If the text has unicode characters, it may not convert the format properly",
    "TXT002": "has multiple text fill colors. Only the first one will be converted",
    "TXT003": "has a non-solid text color (gradient or image) which is not supported by Sketch",
    "TXT004": "contains a 'TITLE' transformation. Sketch does not support this text transformation, so no transformation is applied",
    "TXT005": "contains a LIST style with list markers. This style will be ignored",
    "TXT006": "uses OpenType features not supported by Sketch: {features}. They will be ignored",
    "SHP001": "contains a line with at least one 'Reversed triangle' end. This type of marker does not exist in Sketch. It has been converted to a 'Line' type marker",
    "SHP002": "does not have any points. It will be skipped",
    "STY001": "contains a layer blur and a background blur. Only one will be converted",
    "STY002": "contains a DIAMOND gradient, which is not supported. It is converted to a RADIAL gradient",
    "STY003": "contains a fill with a non-standard blend mode, which is not supported at the fill level (us the layer blend mode instead). It will be ignored",
    # "STY004": "has a cropped image background, which is not supported yet. It will be set to stretch",
    "STY005": "contains a image fill property, which is not supported. The extra properties will be ignored",
    "SYM001": "references an invalid symbol. It will be converted to an empty placeholder group",
    "SYM002": "overrides unsupported properties: {props}. The override will be ignored",
    "SYM003": "overrides unsupported properties: {props}. The instance will be detached",
    "SYM004": "contains an override of an unknown node {node_ref}. The override will be ignored",
    "ART001": "has at least one corner radius which is not supported by sketch artboards. The corner radius will be ignored",
    "ART002": "is being converted to an artboard. However, artboard rotations are not supported. Rotation will be ignored",
    "ART003": "has a style that is not supported by sketch artboards. It will add a background rectangle to the artboard with the frame style",
    "GRP001": "has inner shadows which are not supported at group level in Sketch. It will be copied to the layers inside the frame",
    "GRP002": "has children with different resizing constraints. This cannot be accurately represented in Sketch so the converted group will not scale correctly",
    "CMP001": "uses a shared style which could not be found in the document. It will not be applied",
    "POS001": "contains NaN in the positioning matrix and cannot be converted. It will be skipped",
    "PRT001": "uses an unsupported interaction type: {props}. Prototype interaction ignored",
    "PRT002": "has multiple actions per layer, which is not supported. Only the first one will be converted",
    "PRT003": "has an action with an unsupported navigation type: {props}. This action will be ignored",
    "PRT004": "has an action with an unsupported connection type: {props}. This action will be ignored",
    "PRT005": "has a prototype with a scroll overflow which is not supported. This setting will be ignored.",
    "GRD001": "has a layout grid which is only supported in Sketch artboards. It will be ignored",
    "GRD002": "has multiple grids but their sizes that are not multiples of each other. The larger one will not be converted",
    "GRD003": "has more than three or more grids and Sketch only supports two. Only the two finer grids will be converted",
    "GRD004": "has more than one layouts in the same axis. Only the first one will be converted",
    "GRD005": "has a row layout with a set height which is not supported by Sketch. It will be converted to a layout grid taking the whole height of the artboard",
    "GRD006": "has a row layout with an offset that is not supported. The offset will be ignored",
    "GRD007": "has a row layout with non-integer ratio between gutter and row size. The row layout will not be converted",
    "GRD008": "has a row and column layout. The rows will only be as wide as the columns (instead of taking the entire width of the artboard)",
    "IMG001": "appears to be corrupted in the .fig file, it will not be converted. Try passing --force-convert-images to try to convert it anyway",
    "IMG002": "appears to be corrupted in the .fig file, it will not be converted",
    "IMG003": "is missing from the .fig file, it will not be converted",
    "IMG004": "appears to be corrupted in the .fig file ({error}), it will not be converted",
    "LAY001": "is an unsupported layer type, it will not be converted",
}


def log_conversion_warning(warning_code: str, fig_node: dict, **kw: list) -> None:
    if fig_node["guid"] not in issued_warnings:
        issued_warnings[fig_node["guid"]] = [warning_code]
    elif warning_code not in issued_warnings[fig_node["guid"]]:
        issued_warnings[fig_node["guid"]].append(warning_code)
    else:
        return
    logging.info(
        f"[{warning_code}] {fig_node['type']} '{fig_node['name']}' {WARNING_MESSAGES[warning_code].format(**kw)}"
    )



from .layer_common import AbstractStyledLayer
from .common import Point
from enum import IntEnum
from dataclasses import dataclass, field, InitVar
from typing import List, NamedTuple
import math


class PointRadiusBehaviour(IntEnum):
    V0 = 0
    V1 = 1
    V1_SMOOTH = 2


class CornerStyle(IntEnum):
    ROUNDED = 0
    ROUNDED_INVERTED = 1
    ANGLED = 2
    SQUARED = 3


class CurveMode(IntEnum):
    UNDEFINED = 0
    STRAIGHT = 1
    MIRRORED = 2
    ASYMMETRIC = 3
    DISCONNECTED = 4


@dataclass(kw_only=True)
class CurvePoint:
    _class: str = field(default="curvePoint")
    curveFrom: Point
    curveTo: Point
    point: Point
    cornerRadius: float = 0.0
    cornerStyle: CornerStyle = CornerStyle.ROUNDED
    hasCurveFrom: bool = False
    hasCurveTo: bool = False
    curveMode: CurveMode = CurveMode.UNDEFINED

    @staticmethod
    def Straight(point: Point, radius: float = 0.0) -> "CurvePoint":
        return CurvePoint(
            curveFrom=point,
            curveTo=point,
            point=point,
            cornerRadius=radius,
            curveMode=CurveMode.STRAIGHT,
        )

    @staticmethod
    def Curved(point: Point, curveFrom: Point, curveTo: Point) -> "CurvePoint":
        return CurvePoint(
            curveFrom=curveFrom,
            curveTo=curveTo,
            point=point,
            hasCurveFrom=True,
            hasCurveTo=True,
            curveMode=CurveMode.MIRRORED,
        )


@dataclass(kw_only=True)
class AbstractShapeLayer(AbstractStyledLayer):
    isClosed: bool
    points: List[CurvePoint]
    edited: bool = False
    pointRadiusBehaviour: PointRadiusBehaviour = PointRadiusBehaviour.V1


@dataclass(kw_only=True)
class ShapePath(AbstractShapeLayer):
    _class: str = field(default="shapePath")
    edited: bool = True


@dataclass(kw_only=True)
class Rectangle(AbstractShapeLayer):
    class Corners(NamedTuple):
        topLeft: float
        topRight: float
        bottomRight: float
        bottomLeft: float

    corners: InitVar[Corners] = Corners(0, 0, 0, 0)
    _class: str = field(default="rectangle")
    fixedRadius: float = 0.0
    hasConvertedToNewRoundCorners: bool = True
    needsConvertionToNewRoundCorners: bool = False
    isClosed: bool = True
    points: List[CurvePoint] = field(default_factory=list)

    def __post_init__(self, corners):
        self.points = [
            CurvePoint.Straight(Point(0, 0), corners.topLeft),
            CurvePoint.Straight(Point(0, 1), corners.bottomLeft),
            CurvePoint.Straight(Point(1, 1), corners.bottomRight),
            CurvePoint.Straight(Point(1, 0), corners.topRight),
        ]


def oval_make_points() -> List[CurvePoint]:
    P1 = 0.22385762510000001
    P2 = 0.77614237490000004
    return [
        CurvePoint.Curved(
            point=Point(0.5, 1),
            curveFrom=Point(P2, 1),
            curveTo=Point(P1, 1),
        ),
        CurvePoint.Curved(
            point=Point(1, 0.5),
            curveFrom=Point(1, P1),
            curveTo=Point(1, P2),
        ),
        CurvePoint.Curved(
            point=Point(0.5, 0),
            curveFrom=Point(P1, 0),
            curveTo=Point(P2, 0),
        ),
        CurvePoint.Curved(
            point=Point(0, 0.5),
            curveFrom=Point(0, P2),
            curveTo=Point(0, P1),
        ),
    ]


@dataclass(kw_only=True)
class Oval(AbstractShapeLayer):
    _class: str = field(default="oval")
    points: List[CurvePoint] = field(default_factory=oval_make_points)
    isClosed: bool = True


@dataclass(kw_only=True)
class Star(AbstractShapeLayer):
    _class: str = field(default="star")
    radius: float
    numberOfPoints: float
    isClosed: bool = True
    points: List[CurvePoint] = field(default_factory=list)

    def __post_init__(self):
        for i in range(self.numberOfPoints):
            angle = -math.pi / 2 + 2 * i * math.pi / self.numberOfPoints
            angle2 = angle + math.pi / self.numberOfPoints

            # Outer point
            x1 = 0.5 + (math.cos(angle) * 0.5)
            y1 = 0.5 + (math.sin(angle) * 0.5)
            self.points.append(CurvePoint.Straight(Point(x1, y1)))

            # Inner point shifted by half the angle
            x2 = 0.5 + (math.cos(angle2) * 0.5 * self.radius)
            y2 = 0.5 + (math.sin(angle2) * 0.5 * self.radius)
            self.points.append(CurvePoint.Straight(Point(x2, y2)))


@dataclass(kw_only=True)
class Polygon(AbstractShapeLayer):
    _class: str = field(default="polygon")
    numberOfPoints: float
    points: List[CurvePoint] = field(default_factory=list)
    isClosed: bool = True

    def __post_init__(self):
        for i in range(self.numberOfPoints):
            angle = -math.pi / 2 + 2 * i * math.pi / self.numberOfPoints
            x1 = 0.5 + (math.cos(angle) * 0.5)
            y1 = 0.5 + (math.sin(angle) * 0.5)
            self.points.append(CurvePoint.Straight(Point(x1, y1)))



from sketchformat.layer_group import Page, Artboard, SymbolMaster
from typing import List


def convert(pages: List[Page]) -> dict:
    return {
        "commit": "1899e24f63af087a9dd3c66f73b492b72c27c2c8",
        "pagesAndArtboards": {
            page.do_objectID: {
                "name": page.name,
                "artboards": {
                    artboard.do_objectID: {"name": artboard.name}
                    for artboard in page.layers
                    if isinstance(artboard, Artboard) or isinstance(artboard, SymbolMaster)
                },
            }
            for page in pages
        },
        "version": 144,
        "compatibilityVersion": 99,
        "coeditCompatibilityVersion": 143,
        "app": "com.bohemiancoding.sketch3",
        "autosaved": 0,
        "variant": "NONAPPSTORE",
        "created": {
            "commit": "1899e24f63af087a9dd3c66f73b492b72c27c2c8",
            "appVersion": "93",
            "build": 155335,
            "app": "com.bohemiancoding.sketch3",
            "compatibilityVersion": 99,
            "coeditCompatibilityVersion": 143,
            "version": 144,
            "variant": "NONAPPSTORE",
        },
        "saveHistory": ["NONAPPSTORE.155335"],
        "appVersion": "93",
        "build": 155335,
    }



import logging
from sketchformat.layer_common import *
from sketchformat.layer_shape import *
from sketchformat.style import *
from typing import TypedDict

from . import positioning, style, prototype, utils
from .context import context

SUPPORTED_INHERIT_STYLES = {
    "inheritFillStyleID": ("fillPaints",),
    "inheritFillStyleIDForStroke": (),  # Special cased below
    "inheritStrokeStyleID": (),  # Unused?
    "inheritTextStyleID": (
        "fontName",
        "textCase",
        "fontSize",
        "textDecoration",
        "letterSpacing",
        "lineHeight",
        "paragraphSpacing",
    ),
    "inheritExportStyleID": (),  # Unused?
    "inheritEffectStyleID": ("blur", "shadows", "innerShadows"),
    "inheritGridStyleID": ("layoutGrids"),
    "inheritFillStyleIDForBackground": (),  # Unused?
}


class _Masking(TypedDict):
    shouldBreakMaskChain: bool
    hasClippingMask: bool
    clippingMaskMode: ClippingMaskMode


class _BaseLayer(positioning._Positioning, prototype._Flow):
    do_objectID: str
    name: str
    booleanOperation: BooleanOperation
    exportOptions: ExportOptions
    isFixedToViewport: bool
    isLocked: bool
    isVisible: bool
    layerListExpandedType: LayerListStatus
    nameIsFixed: bool
    resizingConstraint: int
    resizingType: ResizeType
    isTemplate: bool


def base_layer(fig_node: dict) -> _BaseLayer:
    # TODO: Hack for groups that only contain non-visible items
    if math.isnan(fig_node["size"]["x"]):
        fig_node["size"] = {"x": 1, "y": 1}

    return {
        "do_objectID": utils.gen_object_id(fig_node["guid"]),
        "name": fig_node["name"],
        "booleanOperation": BooleanOperation.NONE,
        "exportOptions": export_options(fig_node.get("exportSettings", [])),
        **positioning.convert(fig_node),  # type: ignore
        "isFixedToViewport": False,
        "isLocked": fig_node["locked"],
        "isVisible": fig_node["visible"],
        "layerListExpandedType": LayerListStatus.COLLAPSED,
        "nameIsFixed": False,
        "resizingConstraint": resizing_constraint(fig_node),
        "resizingType": ResizeType.STRETCH,
        **prototype.convert_flow(fig_node),  # type: ignore
        "isTemplate": False,
    }


class _BaseStyled(_BaseLayer, _Masking):
    style: Style


def base_styled(fig_node: dict) -> _BaseStyled:
    obj: _BaseShape = {
        **base_layer(fig_node),  # type: ignore
        **masking(fig_node),  # type: ignore
        "style": process_styles(fig_node),
    }

    if (
        obj["hasClippingMask"]
        and obj["clippingMaskMode"] == ClippingMaskMode.OUTLINE
        and not fig_node.get("f2s_cropped_image")
    ):
        # Outline mask behave differently in fig files and Sketch in regard to fill/stroke colors
        # Remove fill
        obj["style"].fills = []
        # TODO: If we have stroke, we should remove it and enlarge ourselves to occupy that space
        # which is quite tricky in things like shapePaths. This should be pretty rare in practice

    return obj


class _BaseShape(_BaseStyled):
    pointRadiusBehaviour: PointRadiusBehaviour


def base_shape(fig_node: dict) -> _BaseShape:
    return {
        **base_styled(fig_node),  # type: ignore
        # Sketch smooth corners are a boolean, but here it's a percent. Use an arbitrary threshold
        "pointRadiusBehaviour": (
            PointRadiusBehaviour.V1_SMOOTH
            if fig_node.get("cornerSmoothing", 0) > 0.4
            else PointRadiusBehaviour.V1
        ),
    }


def process_styles(fig_node: dict) -> Style:
    # First we apply any overrides that we may have (from inherit* properties)
    # If any of them can be linked as a shared style, we keep track of them to add the IDs at
    # the end
    components = {}
    for inherit_style, copy_keys in SUPPORTED_INHERIT_STYLES.items():
        inherit_node_id = fig_node.get(inherit_style)
        if inherit_node_id is None or utils.is_invalid_ref(inherit_node_id):
            continue

        try:
            inherit_node, sketch_component = context.component(inherit_node_id)
        except KeyError:
            utils.log_conversion_warning("CMP001", fig_node)
            continue

        if inherit_style == "inheritFillStyleIDForStroke":
            fig_node["strokePaints"] = inherit_node["fillPaints"]
        elif inherit_style is not None:
            for key in copy_keys:
                if key in inherit_node:
                    fig_node[key] = inherit_node[key]
        else:
            logging.warning(f"Unsupported {inherit_style}, it will not be copied")

        if sketch_component:
            components[inherit_style] = sketch_component

    st = style.convert(fig_node)

    for key, value in components.items():
        if key == "inheritFillStyleID":
            st.fills[0].color = value.value
        elif key == "inheritFillStyleIDForStroke":
            st.borders[0].color = value.value
        else:
            logging.error(f"Unexpected component for {key}")

    return st


def export_options(fig_export_settings: dict) -> ExportOptions:
    return ExportOptions(
        exportFormats=[
            ExportFormat(
                fileFormat=s["imageType"].lower(),
                name=s["suffix"],
                **export_scale(s["constraint"]),
            )
            for s in fig_export_settings
        ]
    )


class _ExportScale(TypedDict):
    absoluteSize: int
    scale: float
    visibleScaleType: VisibleScaleType


def export_scale(fig_constraint: dict) -> _ExportScale:
    match fig_constraint:
        case {"type": "CONTENT_SCALE", "value": scale}:
            return {
                "absoluteSize": 0,
                "scale": scale,
                "visibleScaleType": VisibleScaleType.SCALE,
            }
        case {"type": "CONTENT_WIDTH", "value": size}:
            return {
                "absoluteSize": size,
                "scale": 0,
                "visibleScaleType": VisibleScaleType.WIDTH,
            }
        case {"type": "CONTENT_HEIGHT", "value": size}:
            return {
                "absoluteSize": size,
                "scale": 0,
                "visibleScaleType": VisibleScaleType.HEIGHT,
            }
        case _:
            logging.warning("Unknown export scale")
            return {
                "absoluteSize": 0,
                "scale": 1,
                "visibleScaleType": VisibleScaleType.SCALE,
            }


# resizingConstraint is a bitfield:
#  1: right sizeable
#  2: width sizeable
#  4: left sizeable
#  8: bottom sizeable
# 16: height sizeable
# 32: top sizeable
# 64: all fixed (should be 0 but it's overridden to mean all sizeable, same as 63). Impossible
HORIZONTAL_CONSTRAINT = {
    "MIN": 1,  # Fixed left + width
    "CENTER": 5,  # Fixed width
    "MAX": 4,  # Fixed right + width
    "STRETCH": 2,  # Fixed left and right
    "SCALE": 7,  # All free
    # 'FIXED_MIN': 0, # Unused?
    # 'FIXED_MAX': 0, # Unused?
}

# Vertical constraints are equivalent to horizontal ones, with a 3 bit shift
VERTICAL_CONSTRAINT = {k: v << 3 for k, v in HORIZONTAL_CONSTRAINT.items()}


def resizing_constraint(fig_node: dict) -> int:
    h = HORIZONTAL_CONSTRAINT[fig_node["horizontalConstraint"]]
    v = VERTICAL_CONSTRAINT[fig_node["verticalConstraint"]]
    return h + v


CLIPPING_MODE = {
    "ALPHA": ClippingMaskMode.ALPHA,
    "OUTLINE": ClippingMaskMode.OUTLINE,
}


def masking(fig_node: dict) -> _Masking:
    return {
        "shouldBreakMaskChain": False,
        "hasClippingMask": bool(fig_node.get("mask")),
        "clippingMaskMode": CLIPPING_MODE[fig_node.get("maskType", "OUTLINE")],
    }



import appdirs
import os
import urllib.request
import urllib.parse
from converter import utils
from fontTools.ttLib import TTFont
from sketchformat.document import FontReference, JsonFileReference
from typing import IO, Tuple
from zipfile import ZipFile

fonts_cache_dir = appdirs.user_cache_dir("Fig2Sketch", "Sketch") + "/fonts"
os.makedirs(fonts_cache_dir, exist_ok=True)


class FontError(Exception):
    pass


def retrieve_webfont(family):
    WEB_FONT_BASE_URL = "http://fonts.google.com/download?family="
    font_url = WEB_FONT_BASE_URL + urllib.parse.quote(family)
    font_file = f"{fonts_cache_dir}/{family}.zip"

    if not os.path.exists(font_file):
        urllib.request.urlretrieve(font_url, font_file)

    return ZipFile(font_file)


def get_webfont(family, subfamily):
    font_zip = retrieve_webfont(family)
    for fi in font_zip.infolist():
        if not fi.filename.endswith((".ttf", ".otf")):
            continue

        font_file = font_zip.open(fi.filename, "r")
        font_names = extract_names(font_file)
        if (
            font_names["family"].lower() == family.lower()
            and font_names["subfamily"].lower() == subfamily.lower()
        ):
            font_file.seek(0)
            return font_file, font_names["postscript"]

    raise FontError(f"Could not find font {family} {subfamily}")


def convert(
    name: Tuple[str, str], font_file: IO[bytes], postscript: str, output_zip: ZipFile
) -> FontReference:
    family, subfamily = name
    data = font_file.read()
    sha = utils.generate_file_ref(data)
    path = f"fonts/{sha}"
    output_zip.open(path, "w").write(data)

    return FontReference(
        do_objectID=utils.gen_object_id((0, 0), bytes.fromhex(sha)),
        fontData=JsonFileReference(_ref_class="MSFontData", _ref=path),
        fontFamilyName=family,
        fontFileName=f"{family}-{subfamily}.ttf",
        postscriptNames=[postscript],
    )


def extract_names(font_file):
    font = TTFont(font_file)
    return {
        "family": font["name"].getBestFamilyName(),
        "subfamily": font["name"].getBestSubFamilyName(),
        "postscript": font["name"].getFirstDebugName((6,)),
    }


def font_matches(fig, font_names):
    return fig["postscript"] == font_names["postscript"] or (
        fig["family"] == font_names["family"] and fig["style"] == font_names["subfamily"]
    )



from .common import Point, WindingRule
from dataclasses import dataclass, field, InitVar
from enum import IntEnum
from typing import Optional, List


class LineCapStyle(IntEnum):
    BUTT = 0
    ROUND = 1
    SQUARE = 2


class LineJoinStyle(IntEnum):
    MITER = 0
    ROUND = 1
    BEVEL = 2


class FillType(IntEnum):
    COLOR = 0
    GRADIENT = 1
    PATTERN = 4


class BorderPosition(IntEnum):
    CENTER = 0
    INSIDE = 1
    OUTSIDE = 2


class BlendMode(IntEnum):
    NORMAL = 0
    DARKEN = 1
    MULTIPLY = 2
    COLOR_BURN = 3
    LIGHTEN = 4
    SCREEN = 5
    COLOR_DODGE = 6
    OVERLAY = 7
    SOFT_LIGHT = 8
    HARD_LIGHT = 9
    DIFFERENCE = 10
    EXCLUSION = 11
    HUE = 12
    SATURATION = 13
    COLOR = 14
    LUMINOSITY = 15
    PLUS_DARKER = 16
    PLUS_LIGHTER = 17


class GradientType(IntEnum):
    LINEAR = 0
    RADIAL = 1
    ANGULAR = 2


class PatternFillType(IntEnum):
    TILE = 0
    FILL = 1
    STRETCH = 2
    FIT = 3


class MarkerType(IntEnum):
    NONE = 0
    OPEN_ARROW = 1
    FILLED_ARROW = 2
    LINE = 3
    OPEN_CIRCLE = 4
    FILLED_CIRCLE = 5
    OPEN_SQUARE = 6
    FILLED_SQUARE = 7


class BlurType(IntEnum):
    GAUSSIAN = 0
    MOTION = 1
    ZOOM = 2
    BACKGROUND = 3


@dataclass(kw_only=True)
class Color:
    _class: str = field(default="color")
    red: float
    green: float
    blue: float
    alpha: float
    swatchID: Optional[str] = None

    @staticmethod
    def Black() -> "Color":
        return Color(red=0, green=0, blue=0, alpha=1)

    @staticmethod
    def White() -> "Color":
        return Color(red=1, green=1, blue=1, alpha=1)

    @staticmethod
    def Translucent() -> "Color":
        return Color(red=0, green=0, blue=0, alpha=0.5)

    @staticmethod
    def Transparent() -> "Color":
        return Color(red=0, green=0, blue=0, alpha=1)

    @staticmethod
    def DefaultFill() -> "Color":
        return Color(red=0.847, green=0.847, blue=0.847, alpha=1)

    @staticmethod
    def DefaultBorder() -> "Color":
        return Color(red=0.592, green=0.592, blue=0.592, alpha=1)


@dataclass(kw_only=True)
class GradientStop:
    _class: str = field(default="gradientStop")
    color: Color
    position: float


@dataclass(kw_only=True)
class Gradient:
    _class: str = field(default="gradient")
    gradientType: GradientType = GradientType.LINEAR
    elipseLength: float = 0
    from_: InitVar[Point] = Point(0.5, 0)
    to: Point = field(default_factory=lambda: Point(0.5, 1))
    stops: List[GradientStop] = field(
        default_factory=lambda: [
            GradientStop(color=Color.White(), position=0),
            GradientStop(color=Color.Black(), position=1),
        ]
    )

    def __post_init__(self, from_):
        setattr(self, "from", from_)

    @staticmethod
    def Linear(from_: Point, to: Point, stops: List[GradientStop]) -> "Gradient":
        return Gradient(gradientType=GradientType.LINEAR, from_=from_, to=to, stops=stops)

    @staticmethod
    def Radial(
        from_: Point, to: Point, elipseLength: float, stops: List[GradientStop]
    ) -> "Gradient":
        return Gradient(
            gradientType=GradientType.RADIAL,
            from_=from_,
            to=to,
            elipseLength=elipseLength,
            stops=stops,
        )

    @staticmethod
    def Angular(stops: List[GradientStop]) -> "Gradient":
        return Gradient(gradientType=GradientType.ANGULAR, stops=stops)


@dataclass(kw_only=True)
class ContextSettings:
    _class: str = field(default="graphicsContextSettings")
    blendMode: BlendMode = BlendMode.NORMAL
    opacity: float = 1


@dataclass(kw_only=True)
class Image:
    _class: str = field(default="MSJSONFileReference")
    _ref_class: str = field(default="MSImageData")
    _ref: str


@dataclass(kw_only=True)
class Fill:
    _class: str = field(default="fill")
    fillType: FillType
    isEnabled: bool = True
    color: Color = field(default_factory=Color.DefaultFill)
    noiseIndex: int = 0
    noiseIntensity: float = 0
    patternFillType: PatternFillType = PatternFillType.TILE
    patternTileScale: float = 1
    contextSettings: ContextSettings = field(default_factory=ContextSettings)
    gradient: Gradient = field(default_factory=Gradient)
    image: Optional[Image] = None

    @staticmethod
    def Color(color: Color, isEnabled: bool = True) -> "Fill":
        return Fill(color=color, fillType=FillType.COLOR, isEnabled=isEnabled)

    @staticmethod
    def Gradient(gradient: Gradient, isEnabled: bool, opacity: float = 1) -> "Fill":
        return Fill(
            gradient=gradient,
            fillType=FillType.GRADIENT,
            isEnabled=isEnabled,
            contextSettings=ContextSettings(opacity=opacity),
        )

    @staticmethod
    def Image(
        path: str,
        patternFillType: PatternFillType,
        patternTileScale: float,
        isEnabled: bool,
        opacity: float = 1,
    ) -> "Fill":
        return Fill(
            image=Image(_ref=path),
            fillType=FillType.PATTERN,
            patternFillType=patternFillType,
            patternTileScale=patternTileScale,
            isEnabled=isEnabled,
            contextSettings=ContextSettings(opacity=opacity),
        )


@dataclass(kw_only=True)
class Border:
    _class: str = field(default="border")
    fillType: FillType
    position: BorderPosition
    thickness: int
    isEnabled: bool = True
    color: Color = field(default_factory=Color.DefaultBorder)
    contextSettings: ContextSettings = field(default_factory=ContextSettings)
    gradient: Gradient = field(default_factory=Gradient)

    @staticmethod
    def from_fill(fill: Fill, position: BorderPosition, thickness: int) -> "Border":
        return Border(
            fillType=fill.fillType,
            color=fill.color,
            gradient=fill.gradient,
            contextSettings=fill.contextSettings,
            position=position,
            thickness=thickness,
            isEnabled=fill.isEnabled,
        )


@dataclass(kw_only=True)
class ColorControls:
    _class: str = field(default="colorControls")
    isEnabled: bool = False
    brightness: float = 0
    contrast: float = 1
    hue: float = 0
    saturation: float = 1


@dataclass(kw_only=True)
class BorderOptions:
    _class: str = field(default="borderOptions")
    isEnabled: bool = True
    lineCapStyle: LineCapStyle = LineCapStyle.BUTT
    lineJoinStyle: LineJoinStyle = LineJoinStyle.MITER
    dashPattern: List[int] = field(default_factory=list)


@dataclass(kw_only=True)
class Blur:
    _class: str = field(default="blur")
    isEnabled: bool = True
    center: Point = field(default_factory=lambda: Point(0.5, 0.5))
    motionAngle: float = 0
    radius: float = 10
    saturation: float = 1
    type: BlurType = BlurType.GAUSSIAN

    @staticmethod
    def Disabled() -> "Blur":
        return Blur(isEnabled=False)


@dataclass(kw_only=True)
class Shadow:
    _class: str = field(default="shadow")
    blurRadius: float
    offsetX: float
    offsetY: float
    spread: float
    isEnabled: bool = True
    color: Color = field(default_factory=Color.Translucent)
    contextSettings: ContextSettings = field(default_factory=ContextSettings)


@dataclass(kw_only=True)
class InnerShadow(Shadow):
    _class: str = field(default="innerShadow")


@dataclass(kw_only=True)
class TextStyle:
    pass


@dataclass(kw_only=True)
class Style:
    _class: str = field(default="style")
    do_objectID: str
    borderOptions: BorderOptions = field(default_factory=BorderOptions)
    borders: List[Border] = field(default_factory=list)
    fills: List[Fill] = field(default_factory=list)
    miterLimit: int = 10
    windingRule: WindingRule = WindingRule.NON_ZERO
    contextSettings: ContextSettings = field(default_factory=ContextSettings)
    colorControls: ColorControls = field(default_factory=ColorControls)
    startMarkerType: MarkerType = MarkerType.NONE
    endMarkerType: MarkerType = MarkerType.NONE
    blur: Blur = field(default_factory=Blur.Disabled)
    textStyle: Optional[TextStyle] = None
    shadows: List[Shadow] = field(default_factory=list)
    innerShadows: List[InnerShadow] = field(default_factory=list)
    startDecorationType: Optional[MarkerType] = None  # Legacy, should match startMarkerType
    endDecorationType: Optional[MarkerType] = None  # Legacy, should match endMarkerType

    def set_markers(self, startMarkerType: MarkerType, endMarkerType: MarkerType) -> None:
        self.startMarkerType = startMarkerType
        self.endMarkerType = endMarkerType

        # Legacy properties, can be skipped. Doing this to match Sketch exactly
        if startMarkerType > 0 or endMarkerType > 0:
            if startMarkerType < 4:
                self.startDecorationType = startMarkerType
            if endMarkerType < 4:
                self.endDecorationType = endMarkerType



import io
import logging
import orjson
from dataclasses import dataclass, field
from typing import Optional, IO


def serialize(obj: object, file: IO[bytes]) -> None:
    file.write(orjson.dumps(obj, default=lambda x: x.to_json()))


# Check if orjson is patched
@dataclass
class Test:
    _class: str = field(default="rectangle")
    optional: Optional[str] = None


b = io.BytesIO()
serialize(Test(), b)
b.seek(0)
if b.read() != b'{"_class":"rectangle"}':
    b.seek(0)
    logging.debug("Unpatched orjson detected. Falling back to built-in json")
    raise Exception()
